from typing import Iterable

from numpy import ndarray

from ..preproc import rhs_disk_mc
from ..protocols import NavierProblemProtocol
from .loads import LoadCase
from ..config import Config

__all__ = ["DiskLoad"]


class DiskLoad(LoadCase[tuple[tuple[float, float], float], Iterable]):
    """
    A class to handle loads defined over a single disk.

    Parameters
    ----------
    domain: tuple[tuple[float, float], float]
        The first value is the coordinates of the center of the disk, the
        second one is the radius.
    value: Iterable
        Load intensities for each dof in the order :math:`f_z, m_x, m_y`.

    .. hint::
        For a detailed explanation of the sign conventions, refer to
        :ref:`this <plate_sign_conventions>` section of the theory guide.

    """

    def rhs(self, problem: NavierProblemProtocol) -> ndarray:
        """
        Returns the coefficients as a NumPy array.

        Parameters
        ----------
        problem: :class:`~sigmaepsilon.solid.fourier.problem.NavierProblem`
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.

        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number
            of harmonic terms involved (defined for the problem).
        """
        assert problem.model_type.is_2d, f"Invalid model type {problem.model_type}."
        n_MC = self._num_mc or Config.NUM_MC_SAMPLES_PLATE
        return rhs_disk_mc(
            problem.size, problem.shape, self.domain, self.value, n_MC=n_MC
        )



from typing import Iterable

import numpy as np
from numpy import ndarray

from ..preproc import rhs_rect_const, rhs_rect_mc
from ..utils import points_to_rectangle_region
from ..protocols import NavierProblemProtocol
from .loads import LoadCase, Float1d, Float2d
from ..config import Config

__all__ = ["RectangleLoad"]


class RectangleLoad(LoadCase[Float2d, Float1d]):
    """
    A class to handle loads defined over a single rectangle.

    Parameters
    ----------
    domain: :class:`~sigmaepsilon.solid.fourier.loads.Float2d`
        The coordinates of the lower-left and upper-right points of the region
        where the load is applied. Default is ``None``.
    value: :class:`~sigmaepsilon.solid.fourier.loads.Float1d`
        Load intensities for each dof in the order :math:`f_z, m_x, m_y`.
    num_mc: int, Optional
        The number of sampling points for Monte Carlo integration. If no value
        is provided, the global config value is used.

    .. hint::
        For a detailed explanation of the sign conventions, refer to
        :ref:`this <plate_sign_conventions>` section of the theory guide.

    """

    @property
    def region(self) -> Iterable:
        """
        Returns the region as a list of 4 values x0, y0, w, and h, where x0 and y0 are
        the coordinates of the bottom-left corner, w and h are the width and height
        of the region.
        """
        return points_to_rectangle_region(np.array(self.domain))

    def rhs(self, problem: NavierProblemProtocol) -> ndarray:
        """
        Returns the coefficients as a NumPy array.

        Parameters
        ----------
        problem: :class:`~sigmaepsilon.solid.fourier.problem.NavierProblem`
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.

        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number
            of harmonic terms involved (defined for the problem).
        """
        assert problem.model_type.is_2d, f"Invalid model type {problem.model_type}."

        domain = np.array(self.domain, dtype=float)
        values = self.value
        n_MC = self._num_mc or Config.NUM_MC_SAMPLES_PLATE

        has_symbolic_load = any(not isinstance(vi, (float, int)) for vi in values)
        if has_symbolic_load:
            return rhs_rect_mc(problem.size, problem.shape, domain, values, n_MC=n_MC)
        else:
            values = np.array(values, dtype=float)
            return rhs_rect_const(problem.size, problem.shape, domain, values)



from abc import abstractmethod

from .result import LoadCaseResultLinStat
from .protocols import LoadGroupProtocol

__all__ = ["NavierProblem"]


class NavierProblem:
    """
    Base class for Navier problems. The sole reason of this class is to
    avoid circular referencing.
    """

    result_class = LoadCaseResultLinStat

    def __init__(self, *, loads: LoadGroupProtocol | None = None):
        self._loads = loads

    @property
    def loads(self) -> LoadGroupProtocol | None:
        """
        Returns the loads.
        """
        return self._loads

    @loads.setter
    def loads(self, value: LoadGroupProtocol | None):
        """
        Sets the loads.
        """
        self._loads = value

    def _postproc_linstat_load_case_result(self, data) -> LoadCaseResultLinStat:
        res = self.result_class(data, name="values")
        return res

    @abstractmethod
    def linear_static_analysis(self, *args, **kwargs):
        pass



from typing import Iterable, Any, Optional, TypeVar, Generic, TypeAlias, Hashable
from types import NoneType
from abc import abstractmethod

from numpy import ndarray
import numpy as np

from sigmaepsilon.deepdict import DeepDict

from ..protocols import NavierProblemProtocol, LoadGroupProtocol, LoadCaseProtocol
from ..postproc import eval_loads_1d, eval_loads_2d

__all__ = ["LoadGroup", "LoadCase"]


Float1d: TypeAlias = Iterable[float]
Float2d: TypeAlias = Iterable[Float1d]

Int1d: TypeAlias = Iterable[int]
Int2d: TypeAlias = Iterable[Int1d]

LoadDomainType = TypeVar("LoadDomainType")
LoadValueType = TypeVar("LoadValueType")


class LoadCase(Generic[LoadDomainType, LoadValueType]):
    """
    Generic base class for all load cases.

    Parameters
    ----------
    domain: LoadDomainType
        The domain of the load.
    value: LoadValueType
        The value of the load.
    num_mc: int, Optional
        The number of sampling points for Monte Carlo integration.
        If no value is provided, the global config value is used.

    """

    def __init__(
        self,
        domain: LoadDomainType,
        value: LoadValueType,
        num_mc: int | NoneType = None,
    ):
        super().__init__()
        self._domain = domain
        self._value = value
        self._num_mc = num_mc

    @property
    def domain(self) -> LoadDomainType:
        """Returns the domain of the load."""
        return self._domain

    @domain.setter
    def domain(self, value: LoadDomainType) -> None:
        """Sets the domain of the load."""
        self._domain = value

    @property
    def value(self) -> LoadValueType:
        """Returns the value of the load."""
        return self._value

    @value.setter
    def value(self, value: LoadValueType) -> None:
        """Sets the value of the load."""
        self._value = value

    @abstractmethod
    def rhs(self, problem: NavierProblemProtocol) -> ndarray:
        raise NotImplementedError("The method 'rhs' must be implemented.")

    def eval_approx(self, problem: NavierProblemProtocol, points: Iterable) -> ndarray:
        """
        Evaluates the Fourier series approximation of the load at the given points.

        The returned array is an 1d array with the same length as the number of points.
        """
        if problem.model_type.is_2d:
            length_X, length_Y = problem.size
            number_of_modes_X, number_of_modes_Y = problem.shape
            rhs = self.rhs(problem)
            points = np.array(points, dtype=float)
            return eval_loads_2d(
                (length_X, length_Y),
                (number_of_modes_X, number_of_modes_Y),
                rhs,
                points,
            )
        elif problem.model_type.is_1d:
            length_X = float(problem.size)
            number_of_modes_X = problem.shape
            rhs = self.rhs(problem)
            points = np.array(points, dtype=float)
            return eval_loads_1d(
                length_X,
                number_of_modes_X,
                rhs,
                points,
            )
        else:  # pragma: no cover
            raise NotImplementedError


class LoadGroup(DeepDict[Hashable, LoadGroupProtocol | LoadCaseProtocol | Any]):
    """
    A class to handle load groups for Navier's semi-analytic solution of
    rectangular plates and beams with specific boundary conditions.

    This class is also the base class of all other load types.

    See Also
    --------
    :class:`~sigmaepsilon.deepdict.deepdict.DeepDict`

    Examples
    --------
    >>> from sigmaepsilon.solid.fourier import LoadGroup, PointLoad
    >>>
    >>> loads = LoadGroup(
    >>>     group1 = LoadGroup(
    >>>         case1 = PointLoad(x=L/3, v=[1.0, 0.0]),
    >>>         case2 = PointLoad(x=L/3, v=[0.0, 1.0]),
    >>>     ),
    >>>     group2 = LoadGroup(
    >>>         case1 = PointLoad(x=2*L/3, v=[1.0, 0.0]),
    >>>         case2 = PointLoad(x=2*L/3, v=[0.0, 1.0]),
    >>>     ),
    >>> )

    Since the LoadGroup class is a subclass of :class:`~sigmaepsilon.deepdict.deepdict.DeepDict`,
    a case is accessible as

    >>> loads['group1', 'case1']

    If you want to protect the object from the accidental
    creation of nested subdirectories, you can lock the layout
    by typing

    >>> loads.lock()
    """

    def __init__(self, *args, cooperative: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self._cooperative = cooperative

    @property
    def cooperative(self) -> bool:
        """
        Returns `True` if the load cases of this group can interact.
        """
        return self._cooperative

    @cooperative.setter
    def cooperative(self, value: bool):
        """
        Sets the cooperativity of the cases in the group.
        """
        self._cooperative = value

    def groups(
        self,
        *,
        inclusive: Optional[bool] = False,
        blocktype: Optional[Any] = None,
        deep: Optional[bool] = True,
    ) -> Iterable[LoadGroupProtocol]:
        """
        Returns a generator object that yields all the subgroups.

        Parameters
        ----------
        inclusive: bool, Optional
            If ``True``, returns the object the call is made upon.
            Default is False.
        blocktype: Any, Optional
            The type of the load groups to return. Default is ``None``, that
            returns all types.
        deep: bool, Optional
            If ``True``, all deep groups are returned separately. Default is ``True``.

        Yields
        ------
        :class:`~sigmaepsilon.solid.fourier.loads.LoadGroup`
        """
        dtype = LoadGroup if blocktype is None else blocktype
        return self.containers(inclusive=inclusive, dtype=dtype, deep=deep)

    def cases(
        self,
        case_type: Any = None,
    ) -> Iterable[LoadCaseProtocol]:
        """
        Returns a generator that yields the load cases in the group.

        Parameters
        ----------
        case_type: Any, Optional
            The type of the load cases to return. Default is None, that
            returns all types.

        Yields
        ------
        :class:`~sigmaepsilon.solid.fourier.loads.LoadCase`
        """
        case_type = LoadCase if case_type is None else case_type
        return filter(
            lambda i: isinstance(i, case_type),
            self.values(deep=True),
        )



from functools import partial

import numpy as np
from numpy import ndarray

from ..preproc import rhs_line_const, rhs_line_1d_mc, rhs_line_2d_mc
from ..protocols import NavierProblemProtocol
from .loads import LoadCase, Float1d, Float2d
from ..utils import is_scalar
from ..config import Config

__all__ = ["LineLoad"]


class LineLoad(LoadCase[Float1d | Float2d, Float1d]):
    """
    A class to handle loads over lines for both beam and plate problems.

    Parameters
    ----------
    domain: :class:`~sigmaepsilon.solid.fourier.loads.Float1d` | :class:`~sigmaepsilon.solid.fourier.loads.Float2d`
        The point of application as an 1d iterable for a beam, a 2d iterable
        for a plate. In the latter case, the first row is the first point, the
        second row is the second point.
    value: :class:`~sigmaepsilon.solid.fourier.loads.Float1d`
        Load intensities for each dof. The order of the dofs for a beam
        is :math:`[F, M]`, for a plate it is :math:`[F, M_x, M_y]`.
    num_mc: int, Optional
        The number of sampling points for Monte Carlo integration.
        If no value is provided, the global config value is used.

    .. hint::
        For a detailed explanation of the sign conventions, refer to
        :ref:`this <sign_conventions>` section of the theory guide.

    """

    def rhs(self, problem: NavierProblemProtocol) -> ndarray:
        """
        Returns the coefficients as a NumPy array.

        Parameters
        ----------
        problem: :class:`~sigmaepsilon.solid.fourier.problem.NavierProblem`
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.

        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number
            of harmonic terms involved (defined for the problem).
        """
        domain = np.array(self.domain, dtype=float)
        values = self.value
        n_MC = self._num_mc or Config.NUM_MC_SAMPLES_PLATE

        if problem.model_type.is_2d:
            assert (
                len(domain.shape) == 2
            ), f"Invalid shape {domain.shape} for the domain."
            assert len(values) == 3, "Invalid shape for load intensities."

            evaluator = partial(rhs_line_2d_mc, n_MC=n_MC)
        elif problem.model_type.is_1d:
            assert (
                len(domain.shape) == 1
            ), f"Invalid shape {domain.shape} for the domain."
            assert (
                len(values) == 2
            ), f"Invalid shape {values.shape} for load intensities."

            if is_scalar(values[0]) and is_scalar(values[1]):
                values = np.array(values, dtype=float)
                evaluator = rhs_line_const
            else:
                evaluator = partial(rhs_line_1d_mc, n_MC=n_MC)
        else:  # pragma: no cover
            raise NotImplementedError

        return evaluator(problem.size, problem.shape, domain, values)



from .loads import LoadCase, LoadGroup
from .lineload import LineLoad
from .pointload import PointLoad
from .rectangleload import RectangleLoad
from .diskload import DiskLoad

__all__ = [
    "LoadCase",
    "LoadGroup",
    "LineLoad",
    "PointLoad",
    "RectangleLoad",
    "DiskLoad",
]



from typing import Iterable
from types import NoneType

import numpy as np
from numpy import sin, cos, ndarray, pi as PI
from numba import njit, prange

from sigmaepsilon.math import atleast1d, atleast2d, atleast3d, atleast4d, itype_of_ftype


__all__ = ["postproc", "postproc_Mindlin", "postproc_Kirchhoff"]


UZ, ROTX, ROTY, CX, CY, CXY, EXZ, EYZ, MX, MY, MXY, QX, QY = list(range(13))


def postproc(
    size: float | Iterable[float],
    shape: int | Iterable[int],
    points: ndarray,
    solution: ndarray,
    loads: ndarray,
    D: float | ndarray,
    S: float | ndarray | NoneType = None,
) -> ndarray:
    """
    Calculates postprocessing items.
    """
    ftype = solution.dtype
    itype = itype_of_ftype(ftype)
    if S is not None:
        if not isinstance(size, Iterable):
            res = postproc_Timoshenko(
                size,
                shape,
                atleast1d(points).astype(ftype),
                atleast3d(solution).astype(ftype),
                atleast1d(D),
                atleast1d(S),
            )
        else:
            res = postproc_Mindlin(
                np.array(size).astype(ftype),
                np.array(shape).astype(itype),
                atleast2d(points).astype(ftype),
                atleast4d(solution).astype(ftype),
                atleast3d(D),
                atleast3d(S),
            )
    else:
        if not isinstance(size, Iterable):
            res = postproc_Bernoulli(
                size,
                shape,
                atleast1d(points).astype(ftype),
                atleast3d(solution).astype(ftype),
                atleast1d(D),
                loads,
            )
        else:
            res = postproc_Kirchhoff(
                np.array(size).astype(ftype),
                np.array(shape).astype(itype),
                atleast2d(points).astype(ftype),
                atleast3d(solution).astype(ftype),
                atleast3d(D),
                loads,
            )
    # (N, nLHS, nRHS, nP, ...)
    res = np.sum(res, axis=0)
    # (nLHS, nRHS, nP, ...)
    return res


@njit(nogil=True, parallel=True, cache=True)
def postproc_Bernoulli(
    L: float, N: int, points: ndarray, solution: ndarray, EI: ndarray, loads
) -> ndarray:
    """
    JIT-compiled function that calculates post-processing quantities
    at selected ponts for multiple left- and right-hand sides.

    Parameters
    ----------
    L: float
        The length of the beam.
    N: int
        The number of harmonic terms.
    points: numpy.ndarray
        1d array of shape (nP,) of coordinates.
    solution: numpy.ndarray
        Results of a Navier solution as a 3d array of shape (nRHS, nLHS, N).
    EI: float
        Bending stiffnesses as a 3d float array of shape (nLHS, 3, 3).

    Returns
    -------
    numpy.ndarray
        5d array of shape (N, nLHS, nRHS, nP, ...) of post-processing items.
        The indices along the last axis denote the following quantities:

            0 : displacement

            1 : rotation

            2 : curvature

            3 : shear strain

            4 : moment

            5 : shear force

    """
    nP = points.shape[0]
    nLHS, nRHS = solution.shape[:2]
    res = np.zeros((N, nLHS, nRHS, nP, 6), dtype=solution.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            for iP in prange(nP):
                x = points[iP]
                for n in prange(1, N + 1):
                    iN = n - 1
                    arg = PI * n / L
                    Sn = sin(x * arg)
                    Cn = cos(x * arg)
                    vn = solution[iLHS, iRHS, iN]
                    q = loads[iRHS, iN, 1]
                    res[iN, iLHS, iRHS, iP, 0] = vn * Sn
                    res[iN, iLHS, iRHS, iP, 1] = vn * arg * Cn
                    res[iN, iLHS, iRHS, iP, 2] = -vn * arg**2 * Sn
                    res[iN, iLHS, iRHS, iP, 4] = -EI[iLHS] * vn * arg**2 * Sn
                    res[iN, iLHS, iRHS, iP, 5] = (EI[iLHS] * vn * arg**3 + q) * Cn
    return res


@njit(nogil=True, parallel=True, cache=True)
def postproc_Timoshenko(
    L: float, N: int, points: ndarray, solution: ndarray, EI: ndarray, GA: ndarray
) -> ndarray:
    """
    JIT-compiled function that calculates post-processing quantities
    at selected ponts for multiple left- and right-hand sides.

    Parameters
    ----------
    L: float
        The length of the beam.
    N: int
        The number of harmonic terms.
    points: numpy.ndarray
        1d array of shape (nP,) of coordinates.
    solution: numpy.ndarray
        Results of a Navier solution as a 4d array of shape (nRHS, nLHS, N, 2).
    EI: float
        Bending stiffnesses as an 1d float array of shape (nLHS).
    GA: float
        Corrected shear stiffnesses as an 1d float array of shape (nLHS).

    Returns
    -------
    numpy.ndarray
        5d array of shape (N, nLHS, nRHS, nP, ...) of post-processing items.
        The indices along the last axis denote the following quantities:

            0 : displacement

            1 : rotation

            2 : curvature

            3 : shear strain

            4 : moment

            5 : shear force

    """
    nP = points.shape[0]
    nLHS, nRHS = solution.shape[:2]
    res = np.zeros((N, nLHS, nRHS, nP, 6), dtype=solution.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            for iP in prange(nP):
                x = points[iP]
                for n in prange(1, N + 1):
                    iN = n - 1
                    arg = PI * n / L
                    Sn = sin(x * arg)
                    Cn = cos(x * arg)
                    vn, rn = solution[iLHS, iRHS, iN]
                    res[iN, iLHS, iRHS, iP, 0] = vn * Sn
                    res[iN, iLHS, iRHS, iP, 1] = rn * Cn
                    res[iN, iLHS, iRHS, iP, 2] = -rn * arg * Sn
                    res[iN, iLHS, iRHS, iP, 3] = (vn * arg - rn) * Cn
                    res[iN, iLHS, iRHS, iP, 4] = -EI[iLHS] * rn * arg * Sn
                    res[iN, iLHS, iRHS, iP, 5] = GA[iLHS] * (vn * arg - rn) * Cn
    return res


@njit(nogil=True, parallel=True, cache=True)
def postproc_Mindlin(
    size, shape: ndarray, points: ndarray, solution: ndarray, D: ndarray, S: ndarray
) -> ndarray:
    """
    JIT-compiled function that calculates post-processing quantities
    at selected ponts for multiple left- and right-hand sides.

    Parameters
    ----------
    size: numpy.ndarray
        Sizes in both directions as an 1d float array of length 2.
    shape: numpy.ndarray
        Number of harmonic terms involved in both directions as an
        1d integer array of length 2.
    points: numpy.ndarray
        2d array of shape (nP, 2) of coordinates.
    solution: numpy.ndarray
        Results of a Navier solution as a 4d array of shape (nRHS, nLHS, M * N, 3).
    D: numpy.ndarray
        Bending stiffnesses as a 3d float array of shape (nLHS, 3, 3).
    S: numpy.ndarray
        Corrected shear stiffness as a 3d float array of shape (nLHS, 2, 2).

    Returns
    -------
    numpy.ndarray
        5d array of shape (M * N, nRHS, nLHS, nP, ...) of post-processing items.
        The indices along the last axis denote the following quantities:

            0 : displacement z

            1 : rotation x

            2 : rotation y

            3 : curvature x

            4 : curvature y

            5 : curvature xy

            6 : shear strain xz

            7 : shear strain yz

            8 : moment x

            9 : moment y

            10 : moment xy

            11 : shear force x

            12 : shear force y

    """
    Lx, Ly = size
    M, N = shape
    nP = points.shape[0]
    nLHS, nRHS = solution.shape[:2]
    res = np.zeros((M * N, nLHS, nRHS, nP, 13), dtype=D.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            D11, D12, D22, D66 = (
                D[iLHS, 0, 0],
                D[iLHS, 0, 1],
                D[iLHS, 1, 1],
                D[iLHS, 2, 2],
            )
            S44, S55 = S[iLHS, 0, 0], S[iLHS, 1, 1]
            for iP in prange(nP):
                xp, yp = points[iP, :2]
                for m in prange(1, M + 1):
                    Sm = sin(PI * m * xp / Lx)
                    Cm = cos(PI * m * xp / Lx)
                    for n in prange(1, N + 1):
                        iMN = (m - 1) * N + n - 1
                        Amn, Bmn, Cmn = solution[iLHS, iRHS, iMN]
                        Sn = sin(PI * n * yp / Ly)
                        Cn = cos(PI * n * yp / Ly)
                        res[iMN, iLHS, iRHS, iP, UZ] = Amn * Sm * Sn
                        res[iMN, iLHS, iRHS, iP, ROTX] = Bmn * Sm * Cn
                        res[iMN, iLHS, iRHS, iP, ROTY] = Cmn * Sn * Cm
                        res[iMN, iLHS, iRHS, iP, CX] = -PI * Cmn * m * Sm * Sn / Lx
                        res[iMN, iLHS, iRHS, iP, CY] = PI * Bmn * n * Sm * Sn / Ly
                        res[iMN, iLHS, iRHS, iP, CXY] = (
                            -PI * Bmn * m * Cm * Cn / Lx + PI * Cmn * n * Cm * Cn / Ly
                        )
                        res[iMN, iLHS, iRHS, iP, EXZ] = (
                            Cmn * Sn * Cm + PI * Amn * m * Sn * Cm / Lx
                        )
                        res[iMN, iLHS, iRHS, iP, EYZ] = (
                            -Bmn * Sm * Cn + PI * Amn * n * Sm * Cn / Ly
                        )
                        res[iMN, iLHS, iRHS, iP, MX] = (
                            PI * Bmn * D12 * n * Sm * Sn / Ly
                            - PI * Cmn * D11 * m * Sm * Sn / Lx
                        )
                        res[iMN, iLHS, iRHS, iP, MY] = (
                            PI * Bmn * D22 * n * Sm * Sn / Ly
                            - PI * Cmn * D12 * m * Sm * Sn / Lx
                        )
                        res[iMN, iLHS, iRHS, iP, MXY] = (
                            -PI * Bmn * D66 * m * Cm * Cn / Lx
                            + PI * Cmn * D66 * n * Cm * Cn / Ly
                        )
                        res[iMN, iLHS, iRHS, iP, QX] = (
                            Cmn * S55 * Sn * Cm + PI * Amn * S55 * m * Sn * Cm / Lx
                        )
                        res[iMN, iLHS, iRHS, iP, QY] = (
                            -Bmn * S44 * Sm * Cn + PI * Amn * S44 * n * Sm * Cn / Ly
                        )
    return res


# fmt: off
@njit(nogil=True, parallel=True, cache=True)
def postproc_Kirchhoff(
    size, shape: ndarray, points: ndarray, solution: ndarray, D: ndarray, loads: ndarray
) -> ndarray:
    """
    JIT-compiled function that calculates post-processing quantities
    at selected ponts for multiple left- and right-hand sides.

    Parameters
    ----------
    size: numpy.ndarray
        Sizes in both directions as an 1d float array of length 2.
    shape: numpy.ndarray
        Number of harmonic terms involved in both directions as an
        1d integer array of length 2.
    points: numpy.ndarray
        2d array of point coordinates of shape (nP, 2).
    solution: numpy.ndarray
        results of a Navier solution as a 3d array of shape (nRHS, nLHS, M * N).
    D: numpy.ndarray
        3d array of bending stiffness terms (nLHS, 3, 3).

    Returns
    -------
    numpy.ndarray[M * N, nRHS, nLHS, nP, ...]
        numpy array of post-processing items. The indices along
        the last axpis denote the following quantities:

            0 : displacement z

            1 : rotation x

            2 : rotation y

            3 : curvature x

            4 : curvature y

            5 : curvature xy

            6 : shear strain xz

            7 : shear strain yz

            8 : moment y

            9 : moment y

            10 : moment xy

            11 : shear force x

            12 : shear force y

    """
    Lx, Ly = size
    M, N = shape
    nP = points.shape[0]
    nLHS, nRHS = solution.shape[:2]
    res = np.zeros((M * N, nLHS, nRHS, nP, 13), dtype=D.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            D11, D12, D22, D66 = (
                D[iLHS, 0, 0],
                D[iLHS, 0, 1],
                D[iLHS, 1, 1],
                D[iLHS, 2, 2],
            )
            for iP in prange(nP):
                x, y = points[iP, :2]
                for m in prange(1, M + 1):
                    Sm = sin(PI * m * x / Lx)
                    Cm = cos(PI * m * x / Lx)
                    for n in prange(1, N + 1):
                        Sn = sin(PI * n * y / Ly)
                        Cn = cos(PI * n * y / Ly)
                        iMN = (m - 1) * N + n - 1
                        Amn = solution[iLHS, iRHS, iMN]
                        qmn, rmn = loads[iRHS, iMN, 1:]
                        res[iMN, iLHS, iRHS, iP, UZ] = Amn * Sm * Sn
                        res[iMN, iLHS, iRHS, iP, ROTX] = PI * Amn * n * Sm * Cn / Ly
                        res[iMN, iLHS, iRHS, iP, ROTY] = -PI * Amn * m * Sn * Cm / Lx
                        res[iMN, iLHS, iRHS, iP, CX] = (
                            PI**2 * Amn * m**2 * Sm * Sn / Lx**2
                        )
                        res[iMN, iLHS, iRHS, iP, CY] = (
                            PI**2 * Amn * n**2 * Sm * Sn / Ly**2
                        )
                        res[iMN, iLHS, iRHS, iP, CXY] = (
                            -2 * PI**2 * Amn * m * n * Cm * Cn / (Lx * Ly)
                        )
                        res[iMN, iLHS, iRHS, iP, MX] = (
                            PI**2 * Amn * D11 * m**2 * Sm * Sn / Lx**2
                            + PI**2 * Amn * D12 * n**2 * Sm * Sn / Ly**2
                        )
                        res[iMN, iLHS, iRHS, iP, MY] = (
                            PI**2 * Amn * D12 * m**2 * Sm * Sn / Lx**2
                            + PI**2 * Amn * D22 * n**2 * Sm * Sn / Ly**2
                        )
                        res[iMN, iLHS, iRHS, iP, MXY] = (
                            -2 * PI**2 * Amn * D66 * m * n * Cm * Cn / (Lx * Ly)
                        )
                        res[iMN, iLHS, iRHS, iP, QX] = (
                            PI**3 * Amn * D11 * m**3 * Sn * Cm / Lx**3
                            + PI**3 * Amn * D12 * m * n**2 * Sn * Cm / (Lx * Ly**2)
                            + 2 * PI**3 * Amn * D66 * m * n**2 * Sn * Cm / (Lx * Ly**2)
                            + rmn * Sn * Cm
                        )
                        res[iMN, iLHS, iRHS, iP, QY] = (
                            PI**3 * Amn * D12 * m**2 * n * Sm * Cn / (Lx**2 * Ly)
                            + PI**3 * Amn * D22 * n**3 * Sm * Cn / Ly**3
                            + 2 * PI**3 * Amn * D66 * m**2 * n * Sm * Cn / (Lx**2 * Ly)
                            - qmn * Sm * Cn
                        )
    return res
# fmt: on

# fmt: off
@njit(nogil=True, parallel=True, cache=True)
def eval_loads_1d(
    size: float,
    shape: int, 
    coeffs:ndarray, 
    points:ndarray,
    ) -> ndarray:
    """
    Evaluates the loads at some points for beam problems.
    """
    Lx = size
    N = shape
    nP = points.shape[0]
    nDOF = 2
    res = np.zeros((nP, nDOF), dtype=float)
    for iP in prange(nP):
        xP = points[iP]
        for n in range(1, N + 1):
            iN = n-1
            res[iP, 0] += (
                coeffs[iN, 0]
                * np.sin(xP * np.pi * n / Lx) 
            )
            res[iP, 1] += (
                coeffs[iN, 1]
                * np.cos(xP * np.pi * n / Lx) 
            )
    return res
# fmt: on

# fmt: off
@njit(nogil=True, parallel=True, cache=True)
def eval_loads_2d(
    size: tuple,
    shape: tuple, 
    coeffs:ndarray, 
    points:ndarray,
    ) -> ndarray:
    """
    Evaluates the loads at some points for plate problems.
    """
    Lx, Ly = size
    M, N = shape
    nP = points.shape[0]
    nDOF = 3
    res = np.zeros((nP, nDOF), dtype=float)
    for iP in prange(nP):
        xP, yP = points[iP]
        for m in range(1, M + 1):
            for n in range(1, N + 1):
                iMN = (m - 1) * N + n - 1
                res[iP, 0] += (
                    coeffs[iMN, 0]
                    * np.sin(xP * np.pi * m / Lx) 
                    * np.sin(yP * np.pi * n / Ly)
                )
                res[iP, 1] += (
                    coeffs[iMN, 1]
                    * np.sin(xP * np.pi * m / Lx) 
                    * np.cos(yP * np.pi * n / Ly)
                )
                res[iP, 2] += (
                    coeffs[iMN, 2]
                    * np.cos(xP * np.pi * m / Lx) 
                    * np.sin(yP * np.pi * n / Ly)
                )
    return res
# fmt: on



from typing import Iterable, Union, Tuple
from types import NoneType
from functools import partial

import numpy as np
from numba import njit, prange
from numpy import ndarray, sin, cos

from sigmaepsilon.math import atleast1d, atleast3d

from .mc import _monte_carlo_1d, _monte_carlo_2d
from .utils import (
    generate_random_points_in_disk,
    generate_random_points_in_rectangle,
    generate_random_points_on_line_segment_1d,
    generate_random_points_on_line_segment_2d,
)


def lhs_Navier(
    size: Union[float, Tuple[float]],
    shape: Union[int, Tuple[int]],
    *,
    D: Union[float, ndarray],
    S: Union[float, ndarray, NoneType] = None,
    **kw,
) -> ndarray:
    """
    Returns coefficient matrices for a Navier solution, for a single or
    multiple left-hand sides.

    Parameters
    ----------
    size: Union[float, Tuple[float]]
        The size of the problem. Scalar for a beam, 2-tuple for a plate.
    shape: Union[int, Tuple[int]]
        The number of harmonic terms used. Scalar for a beam, 2-tuple for a plate.
    D: Union[float, ndarray]
        2d or 3d float array of bending stiffnesses for a plate, scalar or 1d float array
        for a beam.
    S: Union[float, ndarray], Optional
        2d or 3d float array of shear stiffnesses for a plate, scalar or 1d float array
        for a beam. Only for Mindlin-Reissner plates and Euler-Bernoulli beams.
        plates. Default is None.

    Note
    ----
    Shear stiffnesses must include shear correction.

    Returns
    -------
    numpy.ndarray
        The coefficients as an array. See the documentation of the corresponding
        function for further details.

    See Also
    --------
    :func:`lhs_Navier_Mindlin`
    :func:`lhs_Navier_Kirchhoff`
    :func:`lhs_Navier_Bernoulli`
    :func:`lhs_Navier_Timoshenko`
    """
    if isinstance(shape, Iterable):  # plate problem
        if S is None:
            return lhs_Navier_Kirchhoff(size, shape, atleast3d(D))
        else:
            return lhs_Navier_Mindlin(size, shape, atleast3d(D), atleast3d(S))
    else:  # beam problem
        if S is None:
            return lhs_Navier_Bernoulli(size, shape, atleast1d(D))
        else:
            return lhs_Navier_Timoshenko(size, shape, atleast1d(D), atleast1d(S))


@njit(nogil=True, parallel=True, cache=True)
def lhs_Navier_Mindlin(size: tuple, shape: tuple, D: ndarray, S: ndarray) -> ndarray:
    """
    JIT compiled function, that returns coefficient matrices for a Navier
    solution for multiple left-hand sides.

    Parameters
    ----------
    size: tuple
        Tuple of floats, containing the sizes of the rectagle.
    shape: tuple
        Tuple of integers, containing the number of harmonic terms
        included in both directions.
    D: numpy.ndarray
        3d float array of bending stiffnesses.
    S: numpy.ndarray
        3d float array of shear stiffnesses.

    Note
    ----
    The shear stiffness values must include the shear correction factor.

    Returns
    -------
    numpy.ndarray
        4d float array of coefficients.
    """
    Lx, Ly = size
    nLHS = D.shape[0]
    M, N = shape
    PI = np.pi
    res = np.zeros((nLHS, M * N, 3, 3), dtype=D.dtype)
    for iLHS in prange(nLHS):
        D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], D[iLHS, 1, 1], D[iLHS, 2, 2]
        S44, S55 = S[iLHS, 0, 0], S[iLHS, 1, 1]
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                iMN = (m - 1) * N + n - 1
                # sum Fx
                res[iLHS, iMN, 0, 0] = (
                    PI**2 * S44 * n**2 / Ly**2 + PI**2 * S55 * m**2 / Lx**2
                )
                res[iLHS, iMN, 0, 1] = -PI * S44 * n / Ly
                res[iLHS, iMN, 0, 2] = PI * S55 * m / Lx
                # sum Mx
                res[iLHS, iMN, 1, 0] = res[iLHS, iMN, 0, 1]
                res[iLHS, iMN, 1, 1] = (
                    (PI**2) * D22 * n**2 / Ly**2 + PI**2 * D66 * m**2 / Lx**2 + S44
                )
                res[iLHS, iMN, 1, 2] = -(PI**2) * D12 * m * n / (
                    Lx * Ly
                ) - PI**2 * D66 * m * n / (Lx * Ly)
                # sum My
                res[iLHS, iMN, 2, 0] = res[iLHS, iMN, 0, 2]
                res[iLHS, iMN, 2, 1] = res[iLHS, iMN, 1, 2]
                res[iLHS, iMN, 2, 2] = (
                    PI**2 * D11 * m**2 / Lx**2 + PI**2 * D66 * n**2 / Ly**2 + S55
                )
    return res


@njit(nogil=True, parallel=True, cache=True)
def lhs_Navier_Kirchhoff(size: tuple, shape: tuple, D: ndarray) -> ndarray:
    """
    JIT compiled function, that returns coefficient matrices for a Navier
    solution for multiple left-hand sides.

    Parameters
    ----------
    size: tuple
        Tuple of floats, containing the sizes of the rectagle.
    shape: tuple
        Tuple of integers, containing the number of harmonic terms
        included in both directions.
    D: numpy.ndarray
        3d float array of bending stiffnesses.

    Returns
    -------
    numpy.ndarray
        2d float array of coefficients.
    """
    Lx, Ly = size
    nLHS = D.shape[0]
    M, N = shape
    PI = np.pi
    res = np.zeros((nLHS, M * N), dtype=D.dtype)
    for iLHS in prange(nLHS):
        D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], D[iLHS, 1, 1], D[iLHS, 2, 2]
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                iMN = (m - 1) * N + n - 1
                res[iLHS, iMN] = (
                    PI**4 * D11 * m**4 / Lx**4
                    + 2 * PI**4 * D12 * m**2 * n**2 / (Lx**2 * Ly**2)
                    + PI**4 * D22 * n**4 / Ly**4
                    + 4 * PI**4 * D66 * m**2 * n**2 / (Lx**2 * Ly**2)
                )
    return res


@njit(nogil=True, parallel=True, cache=True)
def lhs_Navier_Bernoulli(L: float, N: int, EI: ndarray) -> ndarray:
    """
    JIT compiled function, that returns coefficient matrices for a Navier
    solution for multiple left-hand sides.

    Parameters
    ----------
    L: float
        The length of the beam.
    N: int
        The number of harmonic terms.
    EI: numpy.ndarray
        1d float array of bending stiffnesses.

    Returns
    -------
    numpy.ndarray
        2d float array of coefficients.
    """
    nLHS = EI.shape[0]
    PI = np.pi
    res = np.zeros((nLHS, N), dtype=EI.dtype)
    for iLHS in prange(nLHS):
        for n in prange(1, N + 1):
            res[iLHS, n - 1] = PI**4 * EI[iLHS] * n**4 / L**4
    return res


@njit(nogil=True, parallel=True, cache=True)
def lhs_Navier_Timoshenko(L: float, N: int, EI: ndarray, GA: ndarray) -> ndarray:
    """
    JIT compiled function, that returns coefficient matrices for a Navier
    solution for multiple left-hand sides.

    Parameters
    ----------
    L: float
        The length of the beam.
    N: int
        The number of harmonic terms.
    EI: numpy.ndarray
        1d float array of bending stiffnesses.
    GA: numpy.ndarray
        1d float array of shear stiffnesses.

    Note
    ----
    The shear stiffness values must include the shear correction factor.

    Returns
    -------
    numpy.ndarray
        4d float array of coefficients.
    """
    nLHS = EI.shape[0]
    PI = np.pi
    res = np.zeros((nLHS, N, 2, 2), dtype=EI.dtype)
    for iLHS in prange(nLHS):
        for n in prange(1, N + 1):
            iN = n - 1
            c1 = PI * n / L
            c2 = c1 * PI * n / L
            res[iLHS, iN, 0, 0] = c2 * GA[iLHS]
            res[iLHS, iN, 0, 1] = -c1 * GA[iLHS]
            res[iLHS, iN, 1, 0] = res[iLHS, iN, 0, 1]
            res[iLHS, iN, 1, 1] = GA[iLHS] + c2 * EI[iLHS]
    return res


# RIGHT HAND SIDES


@njit(nogil=True, parallel=True, cache=True)
def rhs_Kirchhoff(coeffs: ndarray, size: tuple) -> ndarray:
    """
    Calculates unknowns for Kirchhoff plates.
    """
    Lx, Ly = size
    nRHS, N = coeffs.shape[:2]
    res = np.zeros((nRHS, N))
    PI = np.pi
    cx = PI / Lx
    cy = PI / Ly
    for i in prange(nRHS):
        for n in prange(N):
            res[i, n] = (
                coeffs[i, n, 0]
                - coeffs[i, n, 1] * cy * (n + 1)
                - coeffs[i, n, 2] * cx * (n + 1)
            )
    return res


@njit(nogil=True, parallel=True, cache=True)
def rhs_Bernoulli(coeffs: ndarray, L: float) -> ndarray:
    """
    Calculates unknowns for Bernoulli beams.
    """
    nRHS, N = coeffs.shape[:2]
    res = np.zeros((nRHS, N))
    c = np.pi / L
    for i in prange(nRHS):
        for n in prange(N):
            res[i, n] = coeffs[i, n, 0] - coeffs[i, n, 1] * c * (n + 1)
    return res


@njit(nogil=True, parallel=True, cache=True)
def _rhs_line_const(L: float, N: int, domain: ndarray, values: ndarray) -> ndarray:
    PI = np.pi
    rhs = np.zeros((N, 2), dtype=domain.dtype)
    for n in prange(1, N + 1):
        iN = n - 1
        xa, xb = domain
        f, m = values
        c = PI * n / L
        rhs[iN, 0] = (2 * f / (PI * n)) * (cos(c * xa) - cos(c * xb))
        rhs[iN, 1] = (2 * m / (PI * n)) * (sin(c * xb) - sin(c * xa))
    return rhs


@njit(nogil=True, cache=True)
def _rhs_rect_const_single_harmonic(
    size: tuple,
    m: int,
    n: int,
    xc: float,
    yc: float,
    w: float,
    h: float,
    values: ndarray,
) -> ndarray:
    Lx, Ly = size
    f, mx, my = values
    PI = np.pi
    return np.array(
        [
            16
            * f
            * sin(PI * m * w / (Lx * 2))
            * sin(PI * m * xc / Lx)
            * sin(PI * h * n / (Ly * 2))
            * sin(PI * n * yc / Ly)
            / (PI**2 * m * n),
            16
            * mx
            * sin(PI * m * w / (Lx * 2))
            * sin(PI * m * xc / Lx)
            * sin(PI * h * n / (Ly * 2))
            * cos(PI * n * yc / Ly)
            / (PI**2 * m * n),
            16
            * my
            * sin(PI * m * w / (Lx * 2))
            * cos(PI * m * xc / Lx)
            * sin(PI * h * n / (Ly * 2))
            * sin(PI * n * yc / Ly)
            / (PI**2 * m * n),
        ]
    )


@njit(nogil=True, parallel=True, cache=True)
def _rhs_rect_const(
    size: tuple, shape: tuple, domain: ndarray, values: ndarray
) -> ndarray:
    M, N = shape
    rhs = np.zeros((M * N, 3), dtype=domain.dtype)
    xmin, ymin = domain[0]
    xmax, ymax = domain[1]
    xc = (xmin + xmax) / 2
    yc = (ymin + ymax) / 2
    w = np.abs(xmax - xmin)
    h = np.abs(ymax - ymin)
    for m in prange(1, M + 1):
        for n in prange(1, N + 1):
            mn = (m - 1) * N + n - 1
            rhs[mn, :] = _rhs_rect_const_single_harmonic(
                size, m, n, xc, yc, w, h, values
            )
    return rhs


@njit(nogil=True, parallel=True, cache=True)
def _rhs_conc_1d(L: tuple, N: tuple, point: float, values: ndarray) -> ndarray:
    c = 2 / L
    rhs = np.zeros((N, 2), dtype=values.dtype)
    PI = np.pi
    f, m = values
    Sx = PI * point / L
    for n in prange(1, N + 1):
        iN = n - 1
        rhs[iN, 0] = c * f * sin(n * Sx)
        rhs[iN, 1] = c * m * cos(n * Sx)
    return rhs


@njit(nogil=True, parallel=True, cache=True)
def _rhs_conc_2d(size: tuple, shape: tuple, point: ndarray, values: ndarray) -> ndarray:
    Lx, Ly = size
    M, N = shape
    rhs = np.zeros((M * N, 3), dtype=point.dtype)
    x, y = point
    fz, mx, my = values
    Sx = np.pi * x / Lx
    Sy = np.pi * y / Ly
    c = 4 / Lx / Ly
    for m in prange(1, M + 1):
        for n in prange(1, N + 1):
            mn = (m - 1) * N + n - 1
            rhs[mn, 0] = c * fz * sin(m * Sx) * sin(n * Sy)
            rhs[mn, 1] = c * mx * cos(m * Sx) * sin(n * Sy)
            rhs[mn, 2] = c * my * sin(m * Sx) * cos(n * Sy)
    return rhs


def rhs_conc_1d(L: float, N: int, domain: float, values: ndarray) -> ndarray:
    """
    Returns coefficients for a concentrated load on a beam.
    Load values are expected in the order [f, m].
    """
    return _rhs_conc_1d(L, N, domain, values)


def rhs_conc_2d(size: tuple, shape: tuple, domain: ndarray, values: ndarray) -> ndarray:
    """
    Returns coefficients for a concentrated load on a plate.
    Load values are expected in the order [f, mx, my].
    """
    return _rhs_conc_2d(size, shape, domain, values)


def rhs_line_const(L: float, N: int, domain: ndarray, values: ndarray) -> ndarray:
    """
    Returns coefficients for constant loads over line segments.
    Values are expected in the order [f, m].
    """
    return _rhs_line_const(L, N, domain, values)


def rhs_line_1d_mc(
    size: float,
    shape: int,
    domain: ndarray,
    values: Iterable,
    *,
    n_MC: int | NoneType = None,
) -> ndarray:
    """
    Returns coefficients for arbitrary loads over line segments.
    Values are expected in the order [f, m].
    """
    length = domain[1] - domain[0]
    rhs = np.zeros((shape, 2), dtype=float)
    rpg = partial(generate_random_points_on_line_segment_1d, domain[0], domain[1])
    _monte_carlo_1d(size, shape, values, length, rpg=rpg, n_MC=n_MC, out=rhs)
    return rhs


def rhs_line_2d_mc(
    size: tuple,
    shape: tuple,
    domain: ndarray,
    values: Iterable,
    *,
    n_MC: int | NoneType = None,
) -> ndarray:
    """
    Returns coefficients for arbitrary line loads for 2d problems.
    Values are expected in the order [f, mx, my].
    """
    length = np.linalg.norm(domain[1] - domain[0])
    rhs = np.zeros((np.prod(shape), 3), dtype=float)
    rpg = partial(generate_random_points_on_line_segment_2d, domain[0], domain[1])
    _monte_carlo_2d(size, shape, values, length, rpg=rpg, n_MC=n_MC, out=rhs)
    return rhs


def rhs_rect_const(
    size: tuple, shape: tuple, domain: ndarray, values: ndarray
) -> ndarray:
    """
    Returns coefficients for constant loads over rectangular patches
    in the order [f, mx, my].
    """
    return _rhs_rect_const(size, shape, domain, values)


def rhs_rect_mc(
    size: tuple,
    shape: tuple,
    domain: ndarray,
    values: Iterable,
    *,
    n_MC: int | NoneType = None,
) -> ndarray:
    """
    Returns coefficients for arbitrary loads over a rectangle.
    Load values are expected in the order [f, mx, my].
    """
    area = np.abs(domain[1, 0] - domain[0, 0]) * np.abs(domain[1, 1] - domain[0, 1])
    rhs = np.zeros((np.prod(shape), 3), dtype=float)
    rpg = partial(generate_random_points_in_rectangle, domain[0], domain[1])
    _monte_carlo_2d(size, shape, values, area, rpg=rpg, n_MC=n_MC, out=rhs)
    return rhs


def rhs_disk_mc(
    size: tuple,
    shape: tuple,
    domain: ndarray,
    values: Iterable,
    *,
    n_MC: int | NoneType = None,
) -> ndarray:
    """
    Returns coefficients for arbitrary loads over a disk.
    Load values are expected in the order [f, mx, my].
    """
    center = domain[0]
    radius = domain[1]
    area = np.pi * radius**2
    rhs = np.zeros((np.prod(shape), 3), dtype=float)
    rpg = partial(generate_random_points_in_disk, center, radius)
    _monte_carlo_2d(size, shape, values, area, rpg=rpg, n_MC=n_MC, out=rhs)
    return rhs



import numpy as np
from numpy import ndarray
from numba import njit, prange

from sigmaepsilon.math.linalg import inv2x2, inv3x3


@njit(nogil=True, parallel=True, cache=True)
def linsolve_Bernoulli(A: ndarray, B: ndarray):
    """
    Calculates unknowns for Bernoulli Beams.
    """
    nLHS, N = A.shape
    nRHS = B.shape[0]
    res = np.zeros((nLHS, nRHS, N))
    for i in prange(nRHS):
        for j in prange(nLHS):
            for n in prange(N):
                res[j, i, n] = B[i, n] / A[j, n]
    return res


@njit(nogil=True, parallel=True, cache=True)
def linsolve_Timoshenko(A: ndarray, B: ndarray):
    """
    Calculates unknowns for Timoshenko Beams.
    """
    nLHS, nMN = A.shape[:2]
    nRHS = B.shape[0]
    res = np.zeros((nLHS, nRHS, nMN, 2))
    for i in prange(nRHS):
        for j in prange(nLHS):
            for k in prange(nMN):
                res[j, i, k] = inv2x2(A[j, k]) @ B[i, k]
    return res


@njit(nogil=True, parallel=True, cache=True)
def linsolve_Mindlin(A: ndarray, B: ndarray):
    nLHS, nMN = A.shape[:2]
    nRHS = B.shape[0]
    res = np.zeros((nLHS, nRHS, nMN, 3))
    for i in prange(nRHS):
        for j in prange(nLHS):
            for k in prange(nMN):
                res[j, i, k] = inv3x3(A[j, k]) @ B[i, k]
    return res


@njit(nogil=True, parallel=True, cache=True)
def linsolve_Kirchhoff(A: ndarray, B: ndarray):
    nLHS, nMN = A.shape[:2]
    nRHS = B.shape[0]
    res = np.zeros((nLHS, nRHS, nMN))
    for i in prange(nRHS):
        for j in prange(nLHS):
            for k in prange(nMN):
                res[j, i, k] = B[i, k] / A[j, k]
    return res



from typing import Iterable, Callable
from types import NoneType

import numpy as np
from numba import njit, prange
from numpy import ndarray

from sigmaepsilon.math.function import Function

from .config import Config


@njit(nogil=True, parallel=True, cache=True)
def _monte_carlo_1d_njit(
    size: float,
    shape: int,
    points: ndarray,
    fvalues: ndarray,
    dsize: float,
    num_MC_tot: int,
    out: ndarray,
) -> ndarray:
    N = shape
    L = size
    multi = (2 / L) * dsize
    factor = np.pi / L

    for n in prange(1, N + 1):
        iN = n - 1
        out[iN, 0] += (
            multi * np.sum(np.sin(points * factor * n) * fvalues[:, 0]) / num_MC_tot
        )
        out[iN, 1] += (
            multi * np.sum(np.cos(points * factor * n) * fvalues[:, 1]) / num_MC_tot
        )

    return out


def _monte_carlo_1d(
    size: float,
    shape: int,
    values: Iterable,
    dsize: float,
    *,
    rpg: Callable[[int], ndarray],
    out: ndarray | NoneType,
    n_MC: int | NoneType = None,
) -> ndarray:
    functions = []
    for i in range(2):
        if isinstance(values[i], str):
            function = Function(values[i], variables=["x"])
        elif isinstance(values[i], (float, int)):

            def function(points) -> ndarray:
                return np.full((len(points),), values[i], dtype=float)

        else:  # pragma: no cover
            raise ValueError(f"Invalid value {values[i]}")
        functions.append(function)

    num_MC_samples = n_MC or Config.NUM_MC_SAMPLES_BEAM
    MC_batch_size = Config.MC_BATCH_SIZE_BEAM
    num_MC_bacthes = int(num_MC_samples // MC_batch_size)
    remaining_MC_points = int(num_MC_samples % MC_batch_size)
    f_vals = np.zeros((MC_batch_size, 2), dtype=float)
    points = np.zeros((MC_batch_size), dtype=float)

    for _ in range(num_MC_bacthes):
        rpg(MC_batch_size, out=points)
        for i in range(2):
            f_vals[:, i] = functions[i]([points])
        _monte_carlo_1d_njit(size, shape, points, f_vals, dsize, num_MC_samples, out)

    MC_batch_size = remaining_MC_points
    if remaining_MC_points > 1:
        points = points[:MC_batch_size]
        f_vals = f_vals[:MC_batch_size]
        rpg(MC_batch_size, out=points)
        for i in range(2):
            f_vals[:, i] = functions[i]([points])
        _monte_carlo_1d_njit(size, shape, points, f_vals, dsize, num_MC_samples, out)

    return out


@njit(nogil=True, parallel=True, cache=True)
def _monte_carlo_2d_njit(
    size: tuple,
    shape: tuple,
    points: ndarray,
    fvalues: ndarray,
    dsize: float,
    num_MC_tot: int,
    out: ndarray,
) -> ndarray:
    M, N = shape
    Lx, Ly = size
    multi = (4 / Lx / Ly) * dsize
    px = points[:, 0]
    py = points[:, 1]
    factor_x = np.pi / Lx
    factor_y = np.pi / Ly

    for m in prange(1, M + 1):
        for n in prange(1, N + 1):
            mn = (m - 1) * N + n - 1
            out[mn, 0] += (
                multi
                * np.sum(
                    np.sin(px * factor_x * m)
                    * np.sin(py * factor_y * n)
                    * fvalues[:, 0]
                )
                / num_MC_tot
            )
            out[mn, 1] += (
                multi
                * np.sum(
                    np.sin(px * factor_x * m)
                    * np.cos(py * factor_y * n)
                    * fvalues[:, 1]
                )
                / num_MC_tot
            )
            out[mn, 2] += (
                multi
                * np.sum(
                    np.cos(px * factor_x * m)
                    * np.sin(py * factor_y * n)
                    * fvalues[:, 2]
                )
                / num_MC_tot
            )

    return out


def _monte_carlo_2d(
    size: tuple,
    shape: tuple,
    values: Iterable,
    dsize: float,
    *,
    rpg: Callable[[int], ndarray],
    out: ndarray | NoneType,
    n_MC: int | NoneType = None,
) -> ndarray:
    functions = []
    for i in range(3):
        if isinstance(values[i], str):
            function = Function(values[i], variables=["x", "y"])
        elif isinstance(values[i], (float, int)):

            def function(points) -> ndarray:
                return np.full((points.shape[1],), values[i], dtype=float)

        else:  # pragma: no cover
            raise ValueError(f"Invalid value {values[i]}")
        functions.append(function)

    num_MC_samples = n_MC or Config.NUM_MC_SAMPLES_PLATE
    MC_batch_size = Config.MC_BATCH_SIZE_PLATE
    num_MC_bacthes = int(num_MC_samples // MC_batch_size)
    remaining_MC_points = int(num_MC_samples % MC_batch_size)
    f_vals = np.zeros((MC_batch_size, 3), dtype=float)
    points = np.zeros((MC_batch_size, 2), dtype=float)

    for _ in range(num_MC_bacthes):
        rpg(MC_batch_size, out=points)
        for i in range(3):
            f_vals[:, i] = functions[i](points.T)
        _monte_carlo_2d_njit(size, shape, points, f_vals, dsize, num_MC_samples, out)

    MC_batch_size = remaining_MC_points
    if remaining_MC_points > 1:
        points = points[:MC_batch_size]
        f_vals = f_vals[:MC_batch_size]
        rpg(MC_batch_size, out=points)
        for i in range(3):
            f_vals[:, i] = functions[i](points.T)
        _monte_carlo_2d_njit(size, shape, points, f_vals, dsize, num_MC_samples, out)

    return out



from types import NoneType

import numpy as np
from numpy import ndarray


def is_scalar(value: any) -> bool:
    """
    Check if a value is a scalar.
    """
    return isinstance(value, (int, float))


def points_to_rectangle_region(points: ndarray) -> tuple[float]:
    xmin = points[:, 0].min()
    ymin = points[:, 1].min()
    xmax = points[:, 0].max()
    ymax = points[:, 1].max()
    return xmin, ymin, xmax - xmin, ymax - ymin


def generate_random_points_in_disk(
    center: tuple[float, float],
    radius: float,
    N: int,
    *,
    out: ndarray | NoneType = None
) -> ndarray:
    """
    Generate N random points uniformly distributed within a disk.
    """
    # Generate N random angles uniformly between 0 and 2*pi
    angles = np.random.uniform(0, 2 * np.pi, N)

    # Generate N random radii uniformly distributed within the disk (use sqrt to ensure uniformity)
    radii = radius * np.sqrt(np.random.uniform(0, 1, N))

    # Calculate x and y coordinates based on polar coordinates (r, theta)
    x_points = center[0] + radii * np.cos(angles)
    y_points = center[1] + radii * np.sin(angles)

    # Combine x and y into a single array of shape (N, 2)
    if out is None:
        out = np.column_stack((x_points, y_points))
    else:
        out[:, 0] = x_points
        out[:, 1] = y_points

    return out


def generate_random_points_in_rectangle(
    bottom_left: tuple[float, float],
    upper_right: tuple[float, float],
    N: int,
    *,
    out: ndarray | NoneType = None
) -> ndarray:
    """
    Generate N random points uniformly distributed within a rectangle.
    """
    # Extract the x and y limits from the bottom-left and upper-right points
    x_min, y_min = bottom_left
    x_max, y_max = upper_right

    # Combine x and y into a single array of shape (N, 2)
    if out is None:
        x_points = np.random.uniform(x_min, x_max, N)
        y_points = np.random.uniform(y_min, y_max, N)
        out = np.column_stack((x_points, y_points))
    else:
        out[:, 0] = np.random.uniform(x_min, x_max, N)
        out[:, 1] = np.random.uniform(y_min, y_max, N)

    return out


def generate_random_points_on_line_segment_1d(
    start: float, end: float, N: int, *, out: ndarray | NoneType = None
) -> ndarray:
    """
    Generate N points uniformly distributed along a line segment in 1d.
    """
    # Generate N random numbers between 0 and 1 to interpolate between start and end
    t = np.random.uniform(0, 1, N)

    # Interpolate between start and end points using the parameter t
    # Combine x and y into a single array of shape (N, 2)
    if out is None:
        out = (1 - t) * start + t * end
    else:
        out[:] = (1 - t) * start + t * end

    return out


def generate_random_points_on_line_segment_2d(
    start: tuple[float, float],
    end: tuple[float, float],
    N: int,
    *,
    out: ndarray | NoneType = None
) -> ndarray:
    """
    Generate N points uniformly distributed along a line segment in 2d.
    """
    # Convert start and end points to numpy arrays
    start = np.array(start)
    end = np.array(end)

    # Generate N random numbers between 0 and 1 to interpolate between start and end
    t = np.random.uniform(0, 1, N)

    # Interpolate between start and end points using the parameter t
    # Combine x and y into a single array of shape (N, 2)
    if out is None:
        out = (1 - t)[:, np.newaxis] * start + t[:, np.newaxis] * end
    else:
        out[:, :] = (1 - t)[:, np.newaxis] * start + t[:, np.newaxis] * end

    return out



#from os.path import dirname, abspath
from importlib.metadata import metadata

#from sigmaepsilon.core.config import namespace_package_name

from .beam import NavierBeam
from .plate import NavierPlate
from .loads import LoadGroup, RectangleLoad, LineLoad, PointLoad, DiskLoad
from .result import BeamLoadCaseResultLinStat, PlateLoadCaseResultLinStat

__all__ = [
    "NavierBeam",
    "NavierPlate",
    "LoadGroup",
    "RectangleLoad",
    "LineLoad",
    "PointLoad",
    "DiskLoad",
    "BeamLoadCaseResultLinStat",
    "PlateLoadCaseResultLinStat",
]

# __pkg_name__ = namespace_package_name(dirname(abspath(__file__)), 10)
__pkg_name__ = "sigmaepsilon.solid.fourier"
__pkg_metadata__ = metadata(__pkg_name__)
__version__ = __pkg_metadata__["version"]
__description__ = __pkg_metadata__["summary"]
del __pkg_metadata__



import numpy as np
from numpy import ndarray

from ..preproc import rhs_conc_1d, rhs_conc_2d
from ..protocols import NavierProblemProtocol
from .loads import LoadCase, Float1d

__all__ = ["PointLoad"]


class PointLoad(LoadCase[float | Float1d, Float1d]):
    """
    A class to handle concentrated loads.

    Parameters
    ----------
    domain: float or :class:`~sigmaepsilon.solid.fourier.loads.Float1d`
        The point of application. A scalar for a beam, an iterable of
        length 2 for a plate.
    value: :class:`~sigmaepsilon.solid.fourier.loads.Float1d`
        Load values for each dof. The order of the dofs for a beam
        is [F, M], for a plate it is [F, Mx, My].

    .. hint::
        For a detailed explanation of the sign conventions, refer to
        :ref:`this <sign_conventions>` section of the theory guide.

    """

    def rhs(self, problem: NavierProblemProtocol) -> ndarray:
        """
        Returns the coefficients as a NumPy array.

        Parameters
        ----------
        problem: :class:`~sigmaepsilon.solid.fourier.problem.NavierProblem`
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is ``None``.

        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number
            of harmonic terms involved (defined for the problem).
        """
        values = np.array(self.value, dtype=float)

        if problem.model_type.is_2d:
            domain = np.array(self.domain, dtype=float)
            evaluator = rhs_conc_2d
        elif problem.model_type.is_1d:
            domain = float(self.domain)
            evaluator = rhs_conc_1d
        else:  # pragma: no cover
            raise NotImplementedError

        return evaluator(problem.size, problem.shape, domain, values)



from typing import Iterable, ClassVar, TYPE_CHECKING, Sequence
import numpy as np


__all__ = [
    "LoadCaseResultLinStat",
    "BeamLoadCaseResultLinStat",
    "PlateLoadCaseResultLinStat",
]


if TYPE_CHECKING:  # pragma: no cover
    import xarray as xr  # Only for type hints
    import pandas as pd  # Only for type hints


class LoadCaseResultLinStat:
    """
    A class to store results of linear static analysis for a single load case.
    """

    __slots__ = ["_data", "_name", "_components"]

    components: ClassVar[Iterable[str]]
    strain_range: ClassVar[Sequence[int]] = []

    def __init__(
        self,
        data,
        *,
        name: str | None = None,
    ):
        self._data = data
        self._name = name

    @property
    def data(self) -> np.ndarray:
        """
        Returns the results as a ``numpy.ndarray``.
        """
        return self._data

    @property
    def values(self) -> np.ndarray:
        """
        Returns the results as a ``numpy.ndarray``.
        """
        return self._data

    @property
    def strains(self) -> np.ndarray:
        """
        Returns the results as a ``numpy.ndarray``.
        """
        if self.strain_range is None:  # pragma: no cover
            raise ValueError("Strain range is not defined.")
        return self._data[:, list(self.strain_range)]

    @property
    def name(self) -> str | None:
        """
        Returns the name of the results.
        """
        return self._name

    @name.setter
    def name(self, value: str | None):
        """
        Sets the name of the results.
        """
        self._name = value

    def to_xarray(self) -> "xr.DataArray":
        """
        Returns the results as an instance of  :class:`xarray.DataArray`.
        """
        import xarray as xr

        nP = len(self.data)
        components = self.components
        coords = [np.arange(nP), components]

        return xr.DataArray(
            self.data, coords=coords, dims=["index", "component"], name=self.name
        )

    def to_pandas(self) -> "pd.DataFrame":
        """
        Returns the results as an instance of :class:`pandas.DataFrame`.
        """
        import pandas as pd

        nP = len(self.data)
        components = self.components
        index = np.arange(nP)
        columns = components

        df = pd.DataFrame(self.data, index=index, columns=columns)
        df.index.name = "index"

        return df


class BeamLoadCaseResultLinStat(LoadCaseResultLinStat):
    """
    A class to store results of linear static analysis for beams and
    a single load case.

    The class is a subclass of the base class :class:`~LoadCaseResultLinStat`, refer to its
    documentation for available methods and properties.

    The underlying data structure is a 2d NumPy array, where the first axis
    goes along the points of evaluation and the second axis goes along the
    following components:

    +-------+--------------------------------------------------+
    | Name  | Description                                      |
    +=======+==================================================+
    | UY    | Displacement in local Y direction                |
    +-------+--------------------------------------------------+
    | ROTZ  | Rotation around local Z axis                     |
    +-------+--------------------------------------------------+
    | CZ    | Curvature related to bending around local Z axis |
    +-------+--------------------------------------------------+
    | EXY   | Shear strain in local Y direction                |
    +-------+--------------------------------------------------+
    | MZ    | Bending moment around local Z axis               |
    +-------+--------------------------------------------------+
    | SY    | Shear force in local Y direction                 |
    +-------+--------------------------------------------------+

    .. hint::
        For a detailed explanation of the sign conventions, refer to
        :ref:`this <beam_sign_conventions>` section of the theory guide.

    See also
    --------
    :class:`~LoadCaseResultLinStat`

    """

    components = [
        "UY",
        "ROTZ",
        "CZ",
        "EXY",
        "MZ",
        "SY",
    ]


class PlateLoadCaseResultLinStat(LoadCaseResultLinStat):
    """
    A class to store results of linear static analysis for plates and
    a single load case.

    The class is a subclass of the base class :class:`~LoadCaseResultLinStat`, refer to its
    documentation for available methods and properties.

    The underlying data structure is a 2d NumPy array, where the first axis
    goes along the points of evaluation and the second axis goes along the
    following components:

    +-------+--------------------------------------------------+
    | Name  | Description                                      |
    +=======+==================================================+
    | UZ    | Displacement in local Z direction                |
    +-------+--------------------------------------------------+
    | ROTX  | Rotation around local X axis (CW)                |
    +-------+--------------------------------------------------+
    | ROTY  | Rotation around local Y axis (CW)                |
    +-------+--------------------------------------------------+
    | CX    | Curvature related to bending around local X axis |
    +-------+--------------------------------------------------+
    | CY    | Curvature related to bending around local Y axis |
    +-------+--------------------------------------------------+
    | CXY   | Twisting curvature                               |
    +-------+--------------------------------------------------+
    | EXZ   | Shear strain in local Y-Z plane                  |
    +-------+--------------------------------------------------+
    | EYZ   | Shear strain in local X-Z plane                  |
    +-------+--------------------------------------------------+
    | MX    | Bending moment around local Y axis (CCW)         |
    +-------+--------------------------------------------------+
    | MY    | Bending moment around local X axis (CW)          |
    +-------+--------------------------------------------------+
    | MXY   | Twisting moment around local Z axis (CW)         |
    +-------+--------------------------------------------------+
    | QX    | Shear force on the local Y-Z plane (+Z)          |
    +-------+--------------------------------------------------+
    | QY    | Shear force on the local X-Z plane  (+Z)         |
    +-------+--------------------------------------------------+

    .. hint::
        For a detailed explanation of the sign conventions, refer to
        :ref:`this <plate_sign_conventions>` section of the theory guide.

    See also
    --------
    :class:`~LoadCaseResultLinStat`

    """

    components = [
        "UZ",
        "ROTX",
        "ROTY",
        "CX",
        "CY",
        "CXY",
        "EXZ",
        "EYZ",
        "MX",
        "MY",
        "MXY",
        "QX",
        "QY",
    ]
    strain_range = range(3, 8)

    @property
    def strains(self) -> np.ndarray:
        """
        Returns the results as a ``numpy.ndarray``.
        """
        # return self._data[:, 3:8]
        return self._data[:, list(self.strain_range)]



from typing import Iterable, Hashable
from types import NoneType

import numpy as np

from sigmaepsilon.deepdict import DeepDict
from sigmaepsilon.math import atleast1d

from .problem import NavierProblem
from .preproc import lhs_Navier, rhs_Bernoulli
from .postproc import postproc
from .proc import linsolve_Bernoulli, linsolve_Timoshenko
from .result import BeamLoadCaseResultLinStat
from .protocols import LoadGroupProtocol
from .enums import MechanicalModelType

__all__ = ["NavierBeam"]


class NavierBeam(NavierProblem):
    """
    A class designed to handle simply-supported plates bent in the X-Y plane
    and solve them using Navier's method. The beam model can be either
    Euler-Bernoulli or Timoshenko, depending on whether shear stiffness is
    provided at instantiation.

    Parameters
    ----------
    length: float
        The length of the beam.
    N: int, Optional
        The number of harmonic terms involved in the approximation.
        Default is 100.
    EI: float
        Bending stiffness.
    GA: float, Optional
        Shear stiffness. Only for Timoshenko beams. Default is None.
    loads: LoadGroup, Optional
        The loads. Default is None.

    Examples
    --------
    To define an Euler-Bernoulli beam of length 10.0 and
    bending stiffness 2000.0 with 100 harmonic terms involved:

    >>> from sigmaepsilon.solid.fourier import NavierBeam
    >>> beam = NavierBeam(10.0, 100, EI=2000.0)

    To define a Timoshenko beam of length 10.0, bending stiffness
    2000.0 and shear stiffness 1500.0 with 100 harmonic terms involved:

    >>> from sigmaepsilon.solid.fourier import NavierBeam
    >>> beam = NavierBeam(10.0, 100, EI=2000.0, GA=1500.0)

    """

    result_class = BeamLoadCaseResultLinStat

    def __init__(
        self,
        length: float,
        N: int = 100,
        *,
        EI: float,
        GA: float | None = None,
        loads: LoadGroupProtocol | None = None,
    ):
        super().__init__(loads=loads)
        self._length = length
        self.EI = EI
        self.GA = GA
        self.N = N

    @property
    def length(self) -> float:
        """The length of the beam."""
        return self._length

    @length.setter
    def length(self, value: float):
        """Sets the length of the beam."""
        self._length = value

    @property
    def size(self) -> float | int:
        """The length of the beam."""
        return self.length

    @size.setter
    def size(self, value: float | int):
        """Sets the length of the beam."""
        self.length = value

    @property
    def shape(self) -> int:
        """The number of harmonic terms involved in the approximation."""
        return self.N

    @shape.setter
    def shape(self, value: int):
        """Sets the number of harmonic terms involved in the approximation."""
        self.N = value

    @property
    def model_type(self) -> MechanicalModelType:
        """The mechanical model type of the beam."""
        if self.GA is None:
            return MechanicalModelType.BERNOULLI_EULER_BEAM
        else:
            return MechanicalModelType.TIMOSHENKO_BEAM

    def linear_static_analysis(
        self,
        *args,
        points: float | Iterable | NoneType = None,
        loads: LoadGroupProtocol | NoneType = None,
    ) -> DeepDict[Hashable, DeepDict | BeamLoadCaseResultLinStat]:
        """
        Performs a linear static analysis and calculates all postprocessing quantities at
        one ore more points.

        Parameters
        ----------
        *args: LoadGroup, float or Iterable
            The loads and the points, in any order.
        loads: :class:`~sigmaepsilon.solid.fourier.loads.LoadGroup`, Optional
            The loads.
        points: float or Iterable, Optional
            A float or an 1d iterable of coordinates, where the results are
            to be evaluated. If it is a scalar, the resulting dictionary
            contains 1d arrays for every quantity, for every load case. If
            there are multiple points, the result attached to a load case is
            a 2d array, where the first axis goes along the points.

        Returns
        -------
        :class:`~sigmaepsilon.deepdict.deepdict.DeepDict`
            A dictionary with the same layout as the loads.

        """
        if len(args) > 0:

            if len(args) > 2:
                raise ValueError("Too many positional arguments.")

            for arg in args:
                if isinstance(arg, LoadGroupProtocol):
                    if loads is not None:
                        raise ValueError("The loads are already provided.")

                    loads = arg
                elif isinstance(arg, (float, Iterable)):
                    if points is not None:
                        raise ValueError("The points are already provided.")

                    points = arg
                else:
                    raise TypeError(f"Invalid argument type {type(arg)}.")

        loads = self.loads if loads is None else loads

        if not isinstance(loads, LoadGroupProtocol):  # pragma: no cover
            raise TypeError("The loads must be an instance of LoadGroup.")

        # STIFFNESS
        lhs = lhs_Navier(self.length, self.N, D=self.EI, S=self.GA)

        # LOADS
        rhs = np.stack(list(lc.rhs(problem=self) for lc in loads.cases()), axis=0)
        # rhs.shape = (nRHS, nMN, nComponent)

        # SOLUTION
        if self.GA is None:
            _RHS = rhs_Bernoulli(rhs, self.length)
            coeffs = linsolve_Bernoulli(lhs, _RHS)
            del _RHS
            # coeffs.shape = (nLHS, nRHS, nMN)
        else:
            coeffs = linsolve_Timoshenko(lhs, rhs)
            # coeffs.shape = (nLHS, nRHS, nMN, 2)

        # POSTPROCESSING
        points = atleast1d(points)
        res = postproc(self.length, self.N, points, coeffs, rhs, self.EI, self.GA)
        # res.shape = (nLHS, nRHS, nPoint, nComponent)

        result = DeepDict()
        for i, (addr, _) in enumerate(loads.items(deep=True, return_address=True)):
            result[addr] = self._postproc_linstat_load_case_result(res[0, i, :, :])
        result.lock()

        return result



from enum import Enum, auto, unique


@unique
class MechanicalModelType(Enum):
    BERNOULLI_EULER_BEAM = auto()
    TIMOSHENKO_BEAM = auto()
    UFLYAND_MINDLIN_PLATE = auto()
    KIRCHHOFF_LOVE_PLATE = auto()

    @property
    def is_1d(self) -> bool:
        return self in {
            MechanicalModelType.BERNOULLI_EULER_BEAM,
            MechanicalModelType.TIMOSHENKO_BEAM,
        }

    @property
    def is_2d(self) -> bool:
        return self in {
            MechanicalModelType.UFLYAND_MINDLIN_PLATE,
            MechanicalModelType.KIRCHHOFF_LOVE_PLATE,
        }



from typing import Protocol, runtime_checkable, Iterable, Hashable, ClassVar
from numpy import ndarray

from sigmaepsilon.deepdict import DeepDict

from .result import LoadCaseResultLinStat
from .enums import MechanicalModelType


@runtime_checkable
class NavierProblemProtocol(Protocol):

    result_class: ClassVar[LoadCaseResultLinStat]

    @property
    def size(self) -> Iterable[float | int] | float | int:
        pass

    @property
    def shape(self, value) -> Iterable[int] | int:
        pass

    @property
    def model_type(self) -> MechanicalModelType:
        pass

    def linear_static_analysis(
        self, points, loads
    ) -> DeepDict[Hashable, DeepDict | LoadCaseResultLinStat]:
        pass


@runtime_checkable
class LoadCaseProtocol(Protocol):

    @property
    def domain(self):
        pass

    @domain.setter
    def domain(self, value):
        pass

    @property
    def value(self):
        pass

    @value.setter
    def value(self, value):
        pass

    def eval_approx(self, problem: NavierProblemProtocol, points: Iterable) -> ndarray:
        """Evaluates the Fourier series approximation of the load at the given points."""
        pass

    def rhs(self, problem: NavierProblemProtocol) -> ndarray:
        """Calculates the Fourier coefficients."""
        pass


@runtime_checkable
class LoadGroupProtocol(Protocol):
    def groups(self, *args, **kwargs) -> Iterable["LoadGroupProtocol"]:
        pass

    def cases(self, *args, **kwargs) -> Iterable[LoadCaseProtocol]:
        pass



from typing import Iterable, Hashable
from types import NoneType

import numpy as np
from numpy import ndarray

from sigmaepsilon.deepdict import DeepDict
from sigmaepsilon.math import atleast2d

from .problem import NavierProblem
from .preproc import lhs_Navier, rhs_Kirchhoff
from .postproc import postproc
from .proc import linsolve_Kirchhoff, linsolve_Mindlin
from .result import PlateLoadCaseResultLinStat
from .protocols import LoadGroupProtocol
from .enums import MechanicalModelType

__all__ = ["NavierPlate"]


class NavierPlate(NavierProblem):
    """
    A class to handle semi-analytic solutions of rectangular plates with
    specific boudary conditions.

    Parameters
    ----------
    size: tuple[float]
        The size of the rectangle.
    shape: tuple[int]
        Numbers of harmonic terms involved in both directions.
    D: Iterable
        3x3 flexural constitutive matrix.
    S: Iterable, Optional
        2x2 shear constitutive matrix.
    loads: :class:`~sigmaepsilon.solid.fourier.loads.LoadGroup`, Optional
        The loads. Default is None.
    """

    result_class = PlateLoadCaseResultLinStat

    def __init__(
        self,
        size: tuple[float],
        shape: tuple[int],
        *,
        D: Iterable,
        S: Iterable | None = None,
        loads: LoadGroupProtocol | None = None,
    ):
        super().__init__(loads=loads)
        self._size = np.array(size, dtype=float)
        self._shape = np.array(shape, dtype=int)
        self.D = np.array(D, dtype=float)
        self.S = None if S is None else np.array(S, dtype=float)

    @property
    def size(self) -> ndarray[float]:
        return self._size

    @property
    def shape(self) -> ndarray[int]:
        return self._shape

    @property
    def model_type(self) -> MechanicalModelType:
        if self.S is None:
            return MechanicalModelType.KIRCHHOFF_LOVE_PLATE
        else:
            return MechanicalModelType.UFLYAND_MINDLIN_PLATE

    def linear_static_analysis(
        self,
        *args,
        points: Iterable | NoneType = None,
        loads: LoadGroupProtocol | NoneType = None,
    ) -> DeepDict[Hashable, DeepDict | PlateLoadCaseResultLinStat]:
        """
        Performs a linear static analysis and calculates all postprocessing quantities at
        one ore more points.

        Parameters
        ----------
        *args: Iterable[float] or :class:`~sigmaepsilon.solid.fourier.loads.LoadGroup`
            The loads and points, in any order.
        loads: :class:`~sigmaepsilon.solid.fourier.loads.LoadGroup`, Optional
            The loads.
        points: Iterable[Iterable[float]], Optional
            2d float array of coordinates, where the results are to be evaluated.

        Returns
        -------
        :class:`~sigmaepsilon.deepdict.deepdict.DeepDict`
            A dictionary with the same layout as the loads.

        """
        if len(args) > 0:

            if len(args) > 2:
                raise ValueError("Too many positional arguments.")

            for arg in args:
                if isinstance(arg, LoadGroupProtocol):
                    if loads is not None:
                        raise ValueError("The loads are already provided.")

                    loads = arg
                elif isinstance(arg, Iterable):
                    if points is not None:
                        raise ValueError("The points are already provided.")

                    points = arg
                else:
                    raise TypeError(f"Invalid argument type {type(arg)}.")

        loads = self.loads if loads is None else loads

        if not isinstance(loads, LoadGroupProtocol):  # pragma: no cover
            raise TypeError("The loads must be an instance of LoadGroup.")

        # STIFFNESS
        lhs = lhs_Navier(self.size, self.shape, D=self.D, S=self.S)

        # LOADS
        rhs = np.stack(list(lc.rhs(problem=self) for lc in loads.cases()), axis=0)
        # rhs.shape = (nRHS, nMN, nComponent)

        # SOLUTION
        if self.S is None:
            _rhs = rhs_Kirchhoff(rhs, self.size)
            coeffs = linsolve_Kirchhoff(lhs, _rhs)
            del _rhs
            # coeffs.shape = (nLHS, nRHS, nMN)
        else:
            coeffs = linsolve_Mindlin(lhs, rhs)
            # coeffs.shape = (nLHS, nRHS, nMN, 3)

        # POSTPROCESSING
        points = atleast2d(points)
        res = postproc(self.size, self.shape, points, coeffs, rhs, self.D, self.S)
        # res.shape = (nLHS, nRHS, nP, nX)

        result = DeepDict()
        for i, (addr, _) in enumerate(loads.items(deep=True, return_address=True)):
            result[addr] = self._postproc_linstat_load_case_result(res[0, i, :, :])
        result.lock()

        return result
